import json
import xml.etree.ElementTree as elementTree

import httpx
import validators

validator_utils = validators


class Validator:
    @staticmethod
    def email(a_value: str):
        return validators.email(a_value)


def is_in_json_format(aText: str):
    try:
        json.loads(aText)
        return True
    except ValueError:
        return False


def is_xml(value):
    try:
        elementTree.fromstring(value)
    except elementTree.ParseError:
        return False
    return True


def is_url_image(image_url):
    image_formats = ("image/png", "image/jpeg", "image/jpg")
    r = httpx.head(image_url)
    if r.headers["content-type"] in image_formats:
        return True
    return False



from abc import ABC, abstractmethod
from datetime import datetime
from typing import List, Optional

from returns.functions import tap
from returns.future import FutureResult
from returns.maybe import Maybe
from returns.pipeline import flow
from returns.pointfree import bind, map_
from returns.result import Result, safe

from dino_seedwork_be.adapters.messaging.rabbitmq import (ConnectionSettings,
                                                          Exchange,
                                                          MessageConsumer,
                                                          MessageListener,
                                                          Queue, Type)
from dino_seedwork_be.utils.functional import feed_args


class ExchangeListener(ABC):
    _message_consumer: Optional[MessageConsumer]
    _queue: Optional[Queue] = None
    _exchange: Exchange
    is_durable: bool = True
    is_queue_durable: bool = True
    is_queue_auto_deleted: bool = False
    is_auto_acknowledged: bool = False
    is_retry: bool = False
    is_exclusive: bool = True

    def __init__(
        self,
        broker_host: str,
        broker_port: int,
        broker_virtual_host: str,
        broker_user_name: Maybe[str],
        broker_user_password: Maybe[str],
    ) -> None:
        self.attach_to_queue(
            broker_host,
            broker_port,
            broker_virtual_host,
            broker_user_name,
            broker_user_password,
        )

    def message_consumer(self) -> Maybe[MessageConsumer]:
        return Maybe.from_optional(self._message_consumer)

    @safe
    def close(self):
        """
        " Closes my queues
        """
        self.queue().map(lambda queue: queue.close())

    def queue(self) -> Maybe[Queue]:
        return Maybe.from_optional(self._queue)

    def exchange(self) -> Exchange:
        return self._exchange

    @abstractmethod
    def exchange_name(self) -> str:
        """
        " Answers the String name of the exchange I listen to.
        " @return str
        """
        pass

    @abstractmethod
    def filtered_dispatch(self, a_type: str, a_text_message: str) -> FutureResult:
        ...

    @abstractmethod
    def listen_to(self) -> List[str]:
        """
        " Answers the kinds of messages I listen to.
        " :return: List[str]
        """
        pass

    @abstractmethod
    def queue_name(self) -> str:
        """
        " Answers the str name of the queue I listen to. By
        " default it is the simple name of my concrete class.
        " May be overridden to change the name.
        " :return: str
        """
        pass

    def set_exchange(self, exchange: Exchange):
        self._exchange = exchange

    def set_queue(self, a_queue: Queue):
        self._queue = a_queue

    def attach_to_queue(
        self,
        broker_host: str,
        broker_port: int,
        broker_virtual_host: str,
        broker_user_name: Maybe[str],
        broker_user_password: Maybe[str],
    ) -> Result:

        """
        " Attaches to the queues I listen to for messages.
        """
        return flow(
            [
                ConnectionSettings.factory(
                    broker_host,
                    broker_port,
                    broker_virtual_host,
                    broker_user_name.value_or(None),
                    broker_user_password.value_or(None),
                ),
                self.exchange_name(),
                self.is_durable,
                lambda exchange: flow(
                    [
                        exchange,
                        self.queue_name(),
                        self.register_consumer,
                        self.is_queue_durable,
                        self.is_queue_auto_deleted,
                        self.is_exclusive,
                    ],
                    feed_args(Queue.factory_exchange_subscriber),
                    map_(tap(self.set_queue)),
                ),
            ],
            feed_args(Exchange.fanout_instance),
            map_(tap(self.set_exchange)),
        )

    def set_message_consumer(self, message_consumer: MessageConsumer):
        self._message_consumer = message_consumer

    def register_consumer(self, queue: Queue) -> Result:
        parent = self

        class MessageListenerAdapter(MessageListener):
            def handle_message(
                self,
                a_type: str,
                a_message_id: str,
                a_time_stamp: datetime,
                a_binary_message: bytes,
                a_delivery_tag: int,
                is_redelivery: bool,
            ) -> FutureResult:
                return parent.filtered_dispatch(a_type, a_binary_message.decode())

        return flow(
            [self.is_auto_acknowledged, queue, self.is_retry],
            feed_args(MessageConsumer.factory),
            map_(tap(self.set_message_consumer)),
            bind(
                lambda msg_consumer: msg_consumer.receive_only(
                    self.listen_to(), MessageListenerAdapter(Type.TEXT)
                )
            ),
        )

    def is_ready_for_consume(self) -> bool:
        return self.message_consumer().map(lambda ms: ms.is_consuming()).value_or(False)

    def run(self):
        self.exchange().run()

    def stop(self):
        self.message_consumer().map(tap(lambda ms: ms.close()))



from dino_seedwork_be.domain import Entity
from dino_seedwork_be.logic import DomainAssertionConcern


class AggregateRoot(Entity, DomainAssertionConcern):
    """Consits of 1+ entities. Spans transaction boundaries."""

    ...



from dataclasses import dataclass, field
from uuid import uuid4

from dino_seedwork_be.domain import ID, IdentifiedDomainObject, ValueObject


@dataclass(frozen=True, kw_only=True)
class IdentifiedValueObject(IdentifiedDomainObject, ValueObject):
    id: ID = field(hash=True, default_factory=lambda: ID(uuid4()))
    """
    Base class for identified objects
    """



from typing import Optional

from validators.btc_address import re


def notEmpty(aValue: str, aMessage: Optional[str] = None):
    if aValue.strip() == "":
        raise ValueError(aMessage or "Value cannot be empty")


def testRegex(aValue: str, aRegex: str, aMessage: Optional[str] = None):
    if re.match(aRegex, aValue) is None:
        raise ValueError(aMessage or "Value is not in valid format")



from dataclasses import dataclass
from typing import Generic, List, TypeVar

ItemType = TypeVar("ItemType")


@dataclass
class PaginationResult(Generic[ItemType]):
    items: List[ItemType]
    total: int
    offset: int
    limit: int


class BaseQuerier:
    pass



import json

from firebase_admin import messaging
from returns.future import FutureFailure, FutureResult, FutureSuccess

from dino_seedwork_be.adapters.messaging.firebase import \
    UserNotificationPushDrivingAdapter

app = messaging.firebase_admin.initialize_app()


class FirebaseNotificationDrivingAdapter(UserNotificationPushDrivingAdapter):
    def push_to_device(
        self, content: dict, type: str, device_token: str
    ) -> FutureResult:
        try:
            message = messaging.Message(
                data={"type": type, "content": json.dumps(content)}, token=device_token
            )

            response = messaging.send(message)

            return FutureSuccess(response)
        except Exception as error:
            return FutureFailure(error)



import traceback

from fastapi import HTTPException, status

from dino_seedwork_be.adapters.logger import SIMPLE_LOGGER
from dino_seedwork_be.adapters.rest import error_detail_with_code
from dino_seedwork_be.exceptions import MainException


def basic_handle_exception(exception):
    match exception:
        case MainException():
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=error_detail_with_code(exception),
            )
        case _:
            SIMPLE_LOGGER.info("Error Occured: %s", exception)
            traceback.print_exc()
            raise exception
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=str(exception),
            )



from abc import ABC, abstractmethod
from typing import List

from sqlalchemy.ext.asyncio.session import AsyncSession

from dino_seedwork_be.storage.uow import DBSessionUser


class ApplicationLifeCycleUsecase(ABC):
    _session: AsyncSession

    def session(self):
        return self._session

    def set_session(self, a_session):
        self._session = a_session

    @abstractmethod
    def get_session_users(self) -> List[DBSessionUser]:
        ...



import uuid
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, List, Optional, TypeVar

from returns.future import FutureResult, FutureSuccess
from returns.maybe import Maybe
from returns.pipeline import flow, pipe
from returns.pointfree import bind, lash

from dino_seedwork_be.domain.value_objects import ID
from dino_seedwork_be.exceptions import NotImplementError
from dino_seedwork_be.storage.uow import DBSessionUser
from dino_seedwork_be.utils.functional import return_v

EntityType = TypeVar("EntityType")
DTOType = TypeVar("DTOType")


@dataclass
class PaginationResultDB(Generic[EntityType]):
    items: List[EntityType]
    limit: int
    size: int
    total: int


class IRepository(DBSessionUser, Generic[EntityType]):
    def get_next_id(self, simple: Optional[bool] = False) -> FutureResult[ID, Any]:
        def check_exist_and_gen() -> FutureResult:
            next_id_candidate = ID(uuid.uuid4())
            return flow(
                next_id_candidate,
                self.get_by_id,
                bind(
                    pipe(
                        bind(return_v(FutureSuccess(next_id_candidate))),
                        lash(lambda _: check_exist_and_gen()),
                    )
                ),
            )

        next_id_candidate = uuid.uuid4()
        match simple:
            case True:
                return FutureSuccess(ID(next_id_candidate))
            case False:
                return check_exist_and_gen()

        return FutureSuccess(ID(next_id_candidate))

    @abstractmethod
    def get_by_id(self, id: ID) -> FutureResult[Maybe[EntityType], Any]:

        raise NotImplementError(
            "Repository does not support the default getById"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    def add(self, entity: EntityType) -> FutureResult:

        raise NotImplementError(
            "Repository does not support the default insert"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    def save(self, entity: EntityType) -> FutureResult:

        raise NotImplementError(
            "Repository does not support the default update"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    def remove(self, id: ID) -> FutureResult:
        raise NotImplementError(
            "Repository does not support the default delete"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    def count(self) -> FutureResult[int, Any]:
        raise NotImplementError(
            "Repository does not support the default count"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    def get_list_pagination(
        self, filter: Any
    ) -> FutureResult[PaginationResultDB[EntityType], Any]:
        raise NotImplementError(
            "Repository does not support the default getListPagination"
            "method, you need to give it the detail logic"
        )



from datetime import datetime
from typing import Optional

from returns.functions import tap
from returns.maybe import Maybe
from returns.pipeline import flow
from returns.pointfree import map_
from returns.result import Result, Success, safe
from toolz.dicttoolz import get_in

from dino_seedwork_be.logic import AssertionConcern
from dino_seedwork_be.serializer import JSONSerializable
from dino_seedwork_be.utils import now_utc, set_protected_attr, unwrap


class DomainEvent(AssertionConcern, JSONSerializable):
    _version: int = 0
    _occurred_on: datetime
    _name: str
    _props: dict = {}

    def __init__(
        self,
        name: str,
        version: int = 0,
        occurred_on: Optional[datetime] = None,
        props: dict = {},
    ):
        unwrap(
            flow(
                None,
                lambda _: self.set_version(version),
                map_(lambda _: self.set_occurred_on(occurred_on or now_utc())),
                map_(lambda _: self.set_name(name)),
                map_(lambda _: self.set_props(props)),
            )
        )

    @staticmethod
    @safe
    def factory(
        occurred_on: datetime,
        name: str,
        version: int = 0,
        props: dict = {},
    ):
        return DomainEvent(name, version, occurred_on, props)

    def occurred_on(self) -> datetime:
        return self._occurred_on

    def version(self) -> int:
        return self._version

    def props(self) -> dict:
        return self._props

    def set_occurred_on(self, a_datetime: datetime) -> Result:
        self._occurred_on = a_datetime
        return Success(None)

    def set_version(self, a_version: int) -> Result:
        self._version = a_version
        return Success(None)

    def set_name(self, a_name: str) -> Result:
        return flow(
            self.assert_argument_not_empty(a_name),
            map_(tap(lambda _: set_protected_attr(self, "_name", a_name))),
        )

    def set_props(self, props: dict) -> Result:
        self._props = props
        return Success(None)

    def get_prop_attr(self, attr_name: str):
        return self._props[attr_name]

    def name(self) -> str:
        return self._name

    def type(self) -> str:
        return self.name()

    def as_dict(self):
        return {
            "version": self.version(),
            "occurred_on": self.occurred_on(),
            "name": self.name(),
            "props": self._props,
        }

    @staticmethod
    def restore(a_dict):
        return DomainEvent(
            version=get_in(["version"], a_dict, 0),
            occurred_on=Maybe.from_optional(get_in(["occurred_on"], a_dict, None))
            .map(str)
            .map(datetime.fromisoformat)
            .value_or(None),
            name=str(get_in(["name"], a_dict)),
            props=get_in(["props"], a_dict, {}),
        )



import json
from typing import Any, Generic, Type, TypeVar

from returns.result import Failure, Result, Success

from dino_seedwork_be.pubsub.Notification import Notification
from dino_seedwork_be.serializer.AbstractSerializer import AbstractSerializer

NotificationT = TypeVar("NotificationT", bound=Notification)


class NotificationSerializer(
    AbstractSerializer,
    Generic[NotificationT],
):
    ins: "NotificationSerializer"

    @staticmethod
    def instance() -> "NotificationSerializer":
        return NotificationSerializer.ins

    def serialize(self, aNotification: NotificationT) -> Result[str, Exception]:
        try:
            return Success(
                "%s"
                % str(self.json_marshaller().encode(aNotification, unpicklable=False))
            )
        except Exception as error:
            return Failure(error)

    def deserialize(self, aJson) -> Result[Notification, Exception]:
        try:
            return Success(Notification.restore(json.loads(aJson)))
        except Exception as error:
            return Failure(error)


NotificationSerializer.ins = NotificationSerializer()



import os
import struct
import zlib
from typing import BinaryIO


def get_image_dimension(file_or_path, close=False):
    """
    Return the (width, height) of an image, given an open file or a path.  Set
    'close' to True to close the file at the end if it is initially in an open
    state.
    """
    from PIL import ImageFile as PillowImageFile

    p = PillowImageFile.Parser()
    if hasattr(file_or_path, "read"):
        file = file_or_path
        file_pos = file.tell()
        file.seek(0)
    else:
        file = open(file_or_path, "rb")
        close = True
    try:
        # Most of the time Pillow only needs a small chunk to parse the image
        # and get the dimensions, but with some TIFF files Pillow needs to
        # parse the whole file.
        chunk_size = 1024
        while 1:
            data = file.read(chunk_size)
            if not data:
                break
            try:
                p.feed(data)
            except zlib.error as e:
                # ignore zlib complaining on truncated stream, just feed more
                # data to parser (ticket #19457).
                if e.args[0].startswith("Error -5"):
                    pass
                else:
                    raise
            except struct.error:
                # Ignore PIL failing on a too short buffer when reads return
                # less bytes than expected. Skip and feed more data to the
                # parser (ticket #24544).
                pass
            except RuntimeError:
                # e.g. "RuntimeError: could not create decoder object" for
                # WebP files. A different chunk_size may work.
                pass
            if p.image:
                return p.image.size
            chunk_size *= 2
        return (None, None)
    finally:
        if close:
            file.close()
        else:
            file.seek(file_pos)


def get_image_file_size(file: BinaryIO) -> float:
    file.seek(0, os.SEEK_END)
    return file.tell()



import datetime
from typing import Generic, Optional, TypeVar, cast

from returns.result import Result, Success

from dino_seedwork_be.domain import IdentifiedDomainObject

from .value_objects import ID, UUID

RawAttributes = TypeVar("RawAttributes")


class Entity(Generic[RawAttributes], IdentifiedDomainObject):
    _created_at: Optional[datetime.datetime] = None
    _updated_at: Optional[datetime.datetime] = None

    _init_param_keys = []

    _concurrency_version: int = 0

    def __init__(self, id: Optional[ID] = None):
        self._concurrency_version = 0
        if id is not None:
            self.setId(id)
        super().__init__()

    @staticmethod
    def init_param_keys():
        return Entity._init_param_keys

    def updated_at(self):
        return self._updated_at

    def created_at(self):
        return self._created_at

    def set_update_at(self, a_date_time: datetime.datetime) -> Result:
        self._updated_at = a_date_time
        return Success("OK")

    def set_created_at(self, a_date_time: datetime.datetime) -> Result:
        self._created_at = a_date_time
        return Success("OK")

    def concurrency_version(self) -> int:
        return self._concurrency_version

    def increase_concurrency_version(self):
        self._concurrency_version += 1

    def __eq__(self, __o) -> bool:
        if __o is None:
            return False
        elif isinstance(__o, Entity):
            return False
        elif self == object:
            return True
        else:
            return self.identity() == cast(Entity, __o).identity()

    @staticmethod
    def create(raw_attributes: RawAttributes, id: UUID):
        print(
            "There is no default method for creation"
            "Please override this create method"
        )



from .BrokerComponent import *
from .ConnectionSettings import *
from .exceptions import *
from .Exchange import *
from .ExchangeListener import *
from .MessageConsumer import *
from .MessageListener import *
from .MessageParameters import *
from .MessageProducer import *
from .Queue import *



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dino_seedwork_be',
 'dino_seedwork_be.adapters',
 'dino_seedwork_be.adapters.messaging',
 'dino_seedwork_be.adapters.messaging.firebase',
 'dino_seedwork_be.adapters.messaging.rabbitmq',
 'dino_seedwork_be.adapters.pubsub',
 'dino_seedwork_be.adapters.rest',
 'dino_seedwork_be.adapters.rest.fastapi',
 'dino_seedwork_be.application',
 'dino_seedwork_be.domain',
 'dino_seedwork_be.event',
 'dino_seedwork_be.event.persistance',
 'dino_seedwork_be.fp',
 'dino_seedwork_be.logic',
 'dino_seedwork_be.media',
 'dino_seedwork_be.process',
 'dino_seedwork_be.pubsub',
 'dino_seedwork_be.serializer',
 'dino_seedwork_be.storage',
 'dino_seedwork_be.storage.alchemysql',
 'dino_seedwork_be.types',
 'dino_seedwork_be.utils',
 'dino_seedwork_be.utils.persistance',
 'dino_seedwork_be.utils.process']

package_data = \
{'': ['*']}

install_requires = \
['Faker>=14.1.0,<15.0.0',
 'Pillow>=9.2.0,<10.0.0',
 'PyJWT[crypto]>=2.4.0,<3.0.0',
 'SQLAlchemy[asyncio]>=1.4.37,<2.0.0',
 'asyncpg>=0.25.0,<0.26.0',
 'fastapi==0.85.0',
 'firebase-admin>=6.0.1,<7.0.0',
 'google-cloud-storage>=2.4.0,<3.0.0',
 'greenlet>=1.1.2,<2.0.0',
 'httpx>=0.23.0,<0.24.0',
 'importlib-resources>=5.7.1,<6.0.0',
 'jsonpickle>=2.2.0,<3.0.0',
 'multimethod>=1.9,<2.0',
 'multipledispatch>=0.6.0,<0.7.0',
 'pika>=1.3.1,<2.0.0',
 'pre-commit>=2.20.0,<3.0.0',
 'psycopg2>=2.9.3,<3.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pyramda>=0.1,<0.2',
 'pytest-asyncio>=0.18.3,<0.19.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'python-multipart>=0.0.5,<0.0.6',
 'pytz>=2022.1,<2023.0',
 'respx>=0.19.2,<0.20.0',
 'returns>=0.19.0,<0.20.0',
 'sqlalchemy-utils>=0.38.3,<0.39.0',
 'toolz>=0.11.2,<0.12.0',
 'uvicorn[standard]>=0.17.6,<0.18.0',
 'validators>=0.19.0,<0.20.0']

setup_kwargs = {
    'name': 'dinohub-seedwork-be',
    'version': '0.1.2',
    'description': 'A seedwork for DinoHub service',
    'long_description': '# DINO hub seedwork\n\nThis is a package contains\n  * DDD (Domain Driven Design) artifact\n  * Functional programming with Returns\n  * Basic helpers function\n',
    'author': 'Tuan Cau Rao',
    'author_email': 'tuan.pt@orai.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import uuid
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, List, Optional, TypeVar

from returns.future import FutureResult

from dino_seedwork_be.domain.value_objects import ID
from dino_seedwork_be.exceptions import NotImplementError
from dino_seedwork_be.storage.uow import DBSessionUser

EntityType = TypeVar("EntityType")
DTOType = TypeVar("DTOType")


@dataclass
class PaginationResultDB(Generic[EntityType]):
    items: List[EntityType]
    limit: int
    size: int
    total: int


class Repository(DBSessionUser, Generic[EntityType]):
    async def get_next_id(self, simple: Optional[bool] = False) -> ID:
        next_id_candidate = uuid.uuid4()
        if simple:
            return ID(next_id_candidate)
        existIdentity = await self.get_by_id(ID(next_id_candidate))
        while existIdentity is not None:
            next_id_candidate = uuid.uuid4()
            existIdentity = await self.get_by_id(ID(next_id_candidate))
        return ID(next_id_candidate)

    @abstractmethod
    async def get_by_id(self, id: ID) -> EntityType:

        raise NotImplementError(
            "Repository does not support the default getById"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    async def add(self, entity: EntityType):

        raise NotImplementError(
            "Repository does not support the default insert"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    async def save(self, entity: EntityType):

        raise NotImplementError(
            "Repository does not support the default update"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    async def remove(self, id: ID):
        raise NotImplementError(
            "Repository does not support the default delete"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    async def count(self):
        raise NotImplementError(
            "Repository does not support the default count"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    async def get_list_pagination(
        self, filter: Any
    ) -> FutureResult[PaginationResultDB[EntityType], Any]:
        raise NotImplementError(
            "Repository does not support the default getListPagination"
            "method, you need to give it the detail logic"
        )
